package org.eclipse.gmt.gems.model.intelligence;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/******************************************************************************
 * Copyright (c) 2007 Jules White.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Jules White - initial API and implementation 
 ****************************************************************************/

/**
 * A ConstraintSet associates constraints with relationships on both
 * specific object instances and ObjectTypes in a knowledge base. Each
 * constraint is keyed by either an ObjectType or object instance and 
 * then a relationship. 
 */
public class ConstraintSet {

	private Map<ObjectType,HashMap<Object, Constraint>> typeConstraints_ = new HashMap<ObjectType, HashMap<Object,Constraint>>();
	private Map<Object, HashMap<Object, Constraint>> instanceConstraints_ = new HashMap<Object, HashMap<Object,Constraint>>();
	
	protected HashMap<Object, List<Constraint>> getConMap(Object key, Map conmap){
		HashMap<Object, List<Constraint>> cons = (HashMap<Object, List<Constraint>>)conmap.get(key);
		if(cons == null){
			cons = new HashMap<Object, List<Constraint>>();
			conmap.put(key, cons);
		}
		return cons;
	}
	
	protected List<Constraint> getList(Object key, HashMap<Object, List<Constraint>> m){
		List<Constraint> l = m.get(key);
		if(l == null){
			l = new ArrayList<Constraint>();
			m.put(key, l);
		}
		return l;
	}
	
	/**
	 * Associates a set of constraints with a specific relationship
	 * of an ObjectType. 
	 * 
	 * @param type the type to associate the constraints with
	 * @param rel the relationship to associate the constraints with
	 * @param cons the constraints to associate
	 */
	public void addAll(ObjectType type, Object rel, Collection<? extends Constraint> cons){
		for(Constraint c : cons){
			add(type,rel,c);
		}
	}
	
	/**
	 * Associates a constraint with a specific relationship
	 * of an ObjectType. 
	 * 
	 * @param type the type to associate the constraints with
	 * @param rel the relationship to associate the constraints with
	 * @param con the constraint to associate
	 */
	public void add(ObjectType type, Object rel, Constraint con){
		getList(rel,getConMap(type, typeConstraints_)).add(con);
	}
	
	/**
	 * Associates a set of constraints with a specific relationship
	 * of an object instance. 
	 * 
	 * @param instance the object to associate the constraints with
	 * @param rel the relationship to associate the constraints with
	 * @param cons the constraints to associate
	 */
	public void addAllToInstance(Object instance, Object rel, Collection<? extends Constraint> cons){
		for(Constraint c : cons){
			addToInstance(instance,rel,c);
		}
	}
	
	/**
	 * Associates a constraint with a specific relationship
	 * of an object instance. 
	 * 
	 * @param instance the object to associate the constraints with
	 * @param rel the relationship to associate the constraints with
	 * @param con the constraint to associate
	 */
	public void addToInstance(Object instance, Object rel, Constraint con){
		getList(rel,getConMap(instance, instanceConstraints_)).add(con);
	}
	
	/**
	 * This method returns all constraints associated with
	 * the specified relationship. The constraints returned
	 * include both the constraints associated specifically
	 * with the object instance and the ObjectType. If only
	 * the instance is specified, the constraints associated
	 * with the type will not be returned. If only the type
	 * is provided, the results will be the constraints 
	 * associated with the relationship for all instances of
	 * that type. The returned constraint sets that include
	 * ObjectType constraints will NOT include any constraints 
	 * associated with the super types of the provided 
	 * ObjectType on that relationship.
	 *  
	 * @param instance the instance that constraints are being 
	 *                 found for (may be null)
	 * @param t the ObjectType that constraints are being found
	 *              found for (may be null if instance is not)
	 * @param rel the relationship that constraints are being
	 *            found for
	 * @return the set of associated constraints
	 */
	public Collection<Constraint> get(Object instance, ObjectType t, Object rel){
		Collection<Constraint> cons = new ArrayList<Constraint>();
		
		if(instance != null){
			cons.addAll(getList(rel,getConMap(instance, instanceConstraints_)));
		}
		if(t != null){
			cons.addAll(get(t, rel));
		}
		
		return cons;
	}
	
	/**
	 * This method returns all constraints associated with
	 * the specified relationship. The returned constraint 
	 * sets that include ObjectType constraints will NOT include 
	 * any constraints associated with the super types of 
	 * the provided ObjectType on that relationship.
	 *  
	 * @param t the ObjectType that constraints are being found
	 *              found for (may be null if instance is not)
	 * @param rel the relationship that constraints are being
	 *            found for
	 * @return the set of associated constraints
	 */
	public Collection<Constraint> get(ObjectType t, Object rel){
		Collection<Constraint> cons = new ArrayList<Constraint>();
		cons.addAll(getList(rel,getConMap(t, typeConstraints_)));
				
		return cons;
	}
	
	/**
	 * This method removes all constraints on the
	 * specified relationship of the provided
	 * ObjectType.
	 * 
	 * @param t the ObjectType containing the
	 *          relationship to remove the constraints
	 *          from
	 * @param rel the relationship to remove the constraints
	 *        from
	 */
	public void remove(ObjectType t, Object rel){
		getList(rel,getConMap(t, typeConstraints_)).clear();
	}
	
	/**
	 * This method removes one constraints on the
	 * specified relationship of the provided
	 * ObjectType.
	 * 
	 * @param t the ObjectType containing the
	 *          relationship to remove the constraints
	 *          from
	 * @param rel the relationship to remove the constraints
	 *        from
	 * @param con the Constraint to remove
	 */
	public void remove(ObjectType t, Object rel, Constraint con){
		getList(rel,getConMap(t, typeConstraints_)).remove(con);
	}
	
	/**
	 * This method removes a constraint that was bound
	 * to a relationship of a specific object instance.
	 * This method will not remove any constraints
	 * associated with the ObjectType of the instance.
	 * 
	 * @param instance the object to remove the constraint from
	 * @param rel the relationship to remove the constraints
	 *        from
	 * @param con the Constraint to remove
	 */
	public void removeFromInstance(Object instance, Object rel, Constraint con){
		getList(rel,getConMap(instance, instanceConstraints_)).remove(con);
	}
}
