package org.eclipse.gmt.gems.model.intelligence;

import java.util.List;

import org.eclipse.gmt.gems.css.parser.Attributes;
import org.eclipse.gmt.gems.css.parser.CSSParser;

/******************************************************************************
 * Copyright (c) 2007 Jules White.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Jules White - initial API and implementation 
 ****************************************************************************/

/**
 * This class contains static utility methods for Constraints and
 * ConstraintSets.
 */
public class Constraints {
	
	/**
	 * This relationship role should be used to specify
	 * traditional ocl-like constraints associated with 
	 * a type rather than a relationship between types.
	 * To check the correctness of a "self" relationship,
	 * call:
	 * <code>
	 * kb.isValid(mySelf, Constraints.SELF_RELATIONSHIP, mySelf)
	 * </code>
	 */
	public static final String SELF_RELATIONSHIP = "Self";

	private static class ObjectTypeNameFunction implements
			Function<ObjectType, String> {
		
		public String apply(ObjectType arg0) {
			return arg0.getName();
		}
	}

	public static final Function<ObjectType, String> OBJECT_TYPE_NAME_FUNCTION = new ObjectTypeNameFunction();

	/**
	 * This is a convenience method for loading a ConstraintSet from a String
	 * specification. The format for the String must be:
	 * 
	 * <code>
	 * ObjectTypeNameRegex {
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   ....
	 * }
	 * 
	 * ObjectTypeNameRegex2 {
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   ....
	 * }
	 * ...
	 * </code>
	 * Example: <code>
	 * Foo {
	 *   Name:ocl{self.name = 'foo'}; 
	 *   Size.*:ocl{self.size.width < 10};
	 * }
	 * .*Bar {
	 *   .*:ocl{self.name = 'bar'}; 
	 *   Size.*:ocl{self.size.width < 20};
	 * }
	 * </code>
	 * 
	 * The example creates the ocl constraint 'self.name = 'foo'' and associates
	 * it with the Name relationship of all Foo type objects. The second
	 * constraint associates the constraint 'self.size.width < 10' with any
	 * relationship of a Foo instance that has a name begining with "Size". The
	 * second group of constraints in the ConstraintSet are for any object with
	 * a type that begins with "Bar". The constraints created match against all
	 * relationship types ".*" and against relationships with names begining
	 * with "Size".
	 * 
	 * @param cons
	 *            the String specification for the ConstraintSet
	 * @return the new ConstraintSet object
	 */
	public static ConstraintSet load(String cons) {
		return load(KBConstraintFactory.getInstance(), cons);
	}

	/**
	 * This is a convenience method for loading a ConstraintSet from a String
	 * specification. The format for the String must be:
	 * 
	 * <code>
	 * ObjectTypeNameRegex {
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   ....
	 * }
	 * 
	 * ObjectTypeNameRegex2 {
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   RelationshipNameRegex:ConstraintType{constraint};
	 *   ....
	 * }
	 * ...
	 * </code>
	 * Example: <code>
	 * Foo {
	 *   Name:ocl{self.name = 'foo'}; 
	 *   Size.*:ocl{self.size.width < 10};
	 * }
	 * .*Bar {
	 *   .*:ocl{self.name = 'bar'}; 
	 *   Size.*:ocl{self.size.width < 20};
	 * }
	 * </code>
	 * 
	 * The example creates the ocl constraint 'self.name = 'foo'' and associates
	 * it with the Name relationship of all Foo type objects. The second
	 * constraint associates the constraint 'self.size.width < 10' with any
	 * relationship of a Foo instance that has a name begining with "Size". The
	 * second group of constraints in the ConstraintSet are for any object with
	 * a type that begins with "Bar". The constraints created match against all
	 * relationship types ".*" and against relationships with names begining
	 * with "Size".
	 * 
	 * The specified KBConstraintFactory is used to construct all Constraint
	 * instances. Each ConstraintType from the String specification is used to
	 * lookup the appropriate ConstraintFactory in the KBConstraintFactory to
	 * delegate the creation call to.
	 * 
	 * @param fact
	 *            the KBConstraintFactory to use to construct Constraint objects
	 * @param cons
	 *            the String specification for the ConstraintSet
	 * @return the new ConstraintSet object
	 */
	public static ConstraintSet load(KBConstraintFactory fact, String cons) {
		List<Attributes> cattrs = CSSParser.parseAttributes(cons, true, false);
		return load(fact,cattrs);
	}
	
	public static ConstraintSet load(List<Attributes> cattrs) {
		return load(KBConstraintFactory.getInstance(),cattrs);
	}
	
	public static Constraint getConstraint(String type, String concode){
		return KBConstraintFactory.getInstance().createConstraint(type, concode);
	}
	
	public static ConstraintSet load(KBConstraintFactory fact, List<Attributes> cattrs) {

		MatchBasedConstraintSet cset = new MatchBasedConstraintSet();
		
		for (Attributes attrs : cattrs) {
			String typesel = attrs.getSelector();
			RegexKBMatcher<ObjectType> matcher = new RegexKBMatcher<ObjectType>(
					typesel, OBJECT_TYPE_NAME_FUNCTION);

			for (Object attr : attrs.keySet()) {
				RegexKBMatcher<Object> relationmatcher = new RegexKBMatcher<Object>(
						"" + attr);
				List values = (List) attrs.get("" + attr);
				for (Object v : values) {
					String rawcode = "" + v;
					String concode = rawcode;
					int split = concode.indexOf("{");
					int end = concode.lastIndexOf("}");
					String type = concode.substring(0, split);
					concode = concode.substring(split + 1, end);
					Constraint con = fact.createConstraint(type, concode);
					
					int msgstart = rawcode.indexOf(",",end);
					if(msgstart != -1){
						String msg = rawcode.substring(msgstart + 1);
						con.setDescription(msg);
					}
					
					if(con == null)
						throw new UnsupportedConstraintTypeException("A constraint could not be created for the type:"+type);
					
					cset.add(matcher, relationmatcher, con);
				}
			}
		}
		return cset;
	}
}
