/*******************************************************************************
 * Copyright (c) 2007 Jules White. All rights reserved. This program and the
 * accompanying materials are made available under the terms of the Eclipse
 * Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: Jules White - initial API and implementation
 ******************************************************************************/
package org.eclipse.gmt.gems.model.intelligence;

import java.util.HashMap;
import java.util.Map;

/**
 * This class is used as a central place for
 * creating Constraint object instances. 
 * The class can be requested to create instances
 * of specific types of constraints. Constraint
 * creation calls are delegated to one of the
 * ConstraintFactories registered with the
 * instance.
 * 
 * @author Jules White
 *
 */
public class KBConstraintFactory {

	
	private static KBConstraintFactory instance_;
	
	/**
	 * This method returns an instance of the
	 * KBFactory. By default, if Prolog, OCL,
	 * Groovy, or BeanShell constraint evaluators
	 * are present in the classpath, they will
	 * be registered with the KBConstraintFactory.
	 * @return the default KBConstraintFactory
	 */
	public static KBConstraintFactory getInstance(){
		if(instance_ == null){
			instance_ = new KBConstraintFactory();
			instance_.register("org.refresh.prolog.PrologConstraint");
			instance_.register("org.eclipse.gmt.gems.model.intelligence.emf.OCLConstraint");
			instance_.register("org.refresh.groovy.GroovyConstraint");
			instance_.register("org.refresh.beanshell.BeanShellConstraint");
		}
		return instance_;
	}
	
	public KBConstraintFactory(){}
	
	private Map<String, ConstraintFactory> factories_ = new HashMap<String, ConstraintFactory>();
	
	/**
	 * This method creates a constraint of the specified type
	 * from the provided data. Internally, the call is delegated
	 * to the ConstraintFactory that is registered with the 
	 * KBFactory for the specified constraint type.
	 * @param type the type of constraint to create
	 * @param data the data to use to construct the constraint (usually a string)
	 * @return the new constraint
	 */
	public Constraint createConstraint(String type, Object data){
		ConstraintFactory fact = factories_.get(type);
		if(fact == null)
			return null;
		
		return fact.createConstraint(data);
	}
	
	/**
	 * This method registers a ConstraintFactory with
	 * the KBFactory. Calls to the KBFactory to create
	 * constraints of the types that correspond to the
	 * ConstraintFactory's getType method, will be
	 * delegated to the ConstraintFactory.
	 * @param f the factory to register
	 */
	public void register(ConstraintFactory f){
		factories_.put(f.getType(),f);
	}
	
	/**
	 * This method registers a ConstraintFactory with
	 * the KBFactory. The method uses Class.forName(fclass).newInstance()
	 * to instantiate the ConstraintFactory. Calls to the KBFactory to create
	 * constraints of the types that correspond to the
	 * ConstraintFactory's getType method, will be
	 * delegated to the ConstraintFactory.
	 * @param f the factory to register
	 */
	public boolean register(String fclass){
		try{
			Class c = Class.forName(fclass);
			ConstraintFactory cf = (ConstraintFactory)c.newInstance();
			register(cf);
		}catch (Throwable e) {
			return false;
		}
		return true;
	}
	
	/**
	 * This method removes a ConstraintFactory so
	 * that it will not be delegated consraint
	 * creation calls from this factory.
	 * @param f the factory to remove
	 */
	public void unRegister(ConstraintFactory f){
		factories_.remove(f.getType());
	}
}
