/*******************************************************************************
 * Copyright (c) 2007 Jules White. All rights reserved. This program and the
 * accompanying materials are made available under the terms of the Eclipse
 * Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: Jules White - initial API and implementation
 ******************************************************************************/
package org.eclipse.gmt.gems.model.intelligence;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * This subclass of ConstraintSet adds the ability to
 * associate constraints with KBMatchers. A KBMatcher
 * typically uses a pattern or expression that is 
 * evaluated against objects to determine if a constraint
 * should be associated with it. 
 * 
 * @author Jules White
 *
 */
public class MatchBasedConstraintSet extends ConstraintSet {

	private class AllMatcher<T> extends KBMatcher<T> {

		
		public boolean matches(T obj) {
			return true;
		}

		
		public Boolean apply(T arg0) {
			return true;
		}

	}

	private class ObjectTypeMatchFunction implements
			Function<ObjectType, Boolean> {
		private ObjectType matchType_;

		public ObjectTypeMatchFunction(ObjectType matchType) {
			super();
			matchType_ = matchType;
		}

		
		public Boolean apply(ObjectType arg0) {
			return ObjectTypes.isAssignableFrom(matchType_, arg0);
		}

	}

	private class ObjectEqualsFunction implements Function<ObjectType, Boolean> {
		private Object ref_;

		public ObjectEqualsFunction(Object ref) {
			super();
			ref_ = ref;
		}

		
		public Boolean apply(ObjectType arg0) {
			return ref_.equals(arg0);
		}

	}

	private List<Constraint> typeConstraints_ = new ArrayList<Constraint>();
	private List<KBMatcher<ObjectType>> typeMatchers_ = new ArrayList<KBMatcher<ObjectType>>();
	private List<KBMatcher<Object>> typeRelationshipMatchers_ = new ArrayList<KBMatcher<Object>>();

	public MatchBasedConstraintSet() {
		super();
	}

	/**
	 * This method adds a constraint that will be associated
	 * with objects that match the specified ObjectType KBMatcher. 
	 * @param typematcher the expression that must match for
	 *                    the constraint to be associated with
	 *                    an object
	 * @param con the constraint to associate
	 */
	public void add(KBMatcher<ObjectType> typematcher, Constraint con) {
		add(typematcher, new AllMatcher<Object>(), con);
	}

	/**
	 * This method adds a constraint that will be associated
	 * with objects of any type that match the specified 
	 * relationship KBMatcher. 
	 * @param relmatcher the expression on the relationship
	 * 					  that must match for
	 *                    the constraint to be associated with
	 *                    an object
	 * @param con the constraint to associate
	 */
	public void addToRelationship(KBMatcher<Object> relmatcher, Constraint con) {
		add(new AllMatcher<ObjectType>(), relmatcher, con);
	}

	/**
	 * This method adds a constraint that will be associated
	 * with the specified relationship for objects that match 
	 * the provided ObjectType KBMatcher. 
	 * @param typematcher the expression that must match for
	 *                    the constraint to be associated with
	 *                    an object
	 * @param rel the relationship to associate the constraint with
	 * @param con the constraint to associate
	 */
	public void add(KBMatcher<ObjectType> typematcher, Object rel,
			Constraint con) {
		add(typematcher, new ObjectEqualsFunction(rel), con);
	}

	/**
	 * This method adds a constraint that will be associated
	 * with objects of the passed type that match the specified 
	 * relationship KBMatcher. 
	 * 
	 * @param type the required type for the object
	 * @param relmatcher the expression on the relationship
	 * 					  that must match for
	 *                    the constraint to be associated with
	 *                    an object
	 * @param con the constraint to associate
	 */
	public void add(ObjectType type, KBMatcher<Object> relationshipmatcher,
			Constraint con) {
		add(new KBMatcher<ObjectType>(new ObjectTypeMatchFunction(type)),
				relationshipmatcher, con);
	}

	/**
	 * This method adds a constraint that will be associated
	 * with any relationship matched by the relationshipmatcher
	 * and for any objects that match the provided ObjectType KBMatcher. 
	 * @param typematcher the expression that must match for
	 *                    the constraint to be associated with
	 *                    an object
	 * @param relaionshipmatcher the relationship matcher to use to
	 *                           associate the constraint with
	 * @param con the constraint to associate
	 */
	public void add(KBMatcher<ObjectType> typematcher,
			KBMatcher<Object> relationshipmatcher, Constraint con) {

		typeConstraints_.add(con);
		typeMatchers_.add(typematcher);
		typeRelationshipMatchers_.add(relationshipmatcher);
	}

	
	
	public Collection<Constraint> get(ObjectType t, Object rel) {
		Collection<Constraint> result = super.get(t, rel);

		for (int i = 0; i < typeConstraints_.size(); i++) {
			if (typeMatchers_.get(i).matches(t)
					&& typeRelationshipMatchers_.get(i).matches(rel)) {
				result.add(typeConstraints_.get(i));
			}
		}

		return result;
	}
	
	/**
	 * This method removes all constraints on the
	 * specified relationship of the provided
	 * ObjectType. It will not remove any 
	 * constraints that have been bound via
	 * matchers.
	 * 
	 * @param t the ObjectType containing the
	 *          relationship to remove the constraints
	 *          from
	 * @param rel the relationship to remove the constraints
	 *        from
	 */
	public void remove(ObjectType t, Object rel){
		super.remove(t, rel);
	}
	
	/**
	 * This method removes one constraint on the
	 * specified relationship of the provided
	 * ObjectType. This method will remove any
	 * constraints bound to matchers that are
	 * identical.
	 * 
	 * @param t the ObjectType containing the
	 *          relationship to remove the constraints
	 *          from
	 * @param rel the relationship to remove the constraints
	 *        from
	 * @param con the Constraint to remove
	 */
	public void remove(ObjectType t, Object rel, Constraint con){
		for (int i = 0; i < typeConstraints_.size(); i++) {
			if (typeMatchers_.get(i).matches(t)
					&& typeRelationshipMatchers_.get(i).matches(rel)) {
				typeConstraints_.remove(i);
				typeMatchers_.remove(i);
				typeRelationshipMatchers_.remove(i);
				break;
			}
		}
		
		super.remove(t, rel, con);
	}
	
	
}
