package org.eclipse.gmt.gems.model.intelligence;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/*******************************************************************************
 * Copyright (c) 2007 Jules White. All rights reserved. This program and the
 * accompanying materials are made available under the terms of the Eclipse
 * Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: Jules White - initial API and implementation
 ******************************************************************************/

/**
 * This class is the core of the Model Intelligence libraries. Objects should be
 * asserted into the knowledge base using the "add" methods. Constraints are
 * then added. Finally, the kb is used to query for endpoints of relationships.
 */
public class RefreshKB implements ConstraintEvaluatorProvider {

	public class UnsupportedObjectType extends RuntimeException {
		public UnsupportedObjectType(String msg) {
			super(msg);
		}
	}

	public class UnsupportedConstraintType extends RuntimeException {
		public UnsupportedConstraintType(String msg) {
			super(msg);
		}
	}

	private List<KBManager> knowledgeManagers_ = new ArrayList<KBManager>();

	private Map<String, ConstraintEvaluator> constraintEvaluators_ = new HashMap<String, ConstraintEvaluator>();

	private Map<String, QueryEvaluator> queryEvaluators_ = new HashMap<String, QueryEvaluator>();

	private List<ObjectView> objectViews_ = new ArrayList<ObjectView>();

	private List<ObjectType> objectTypes_ = new ArrayList<ObjectType>();

	private Map<Object, ObjectView> objectViewLookup_ = new HashMap<Object, ObjectView>();

	private Map<Object, List<ObjectType>> objectTypeLookup_ = new HashMap<Object, List<ObjectType>>();

	private Map<ObjectType, Map<Object, List<ExecutableQuery>>> queryLookup_ = new HashMap<ObjectType, Map<Object, List<ExecutableQuery>>>();

	private Map<ObjectType, List<Object>> objectsByType_ = new HashMap<ObjectType, List<Object>>();

	private List<ObjectTypeSystem> typeSystems_ = new ArrayList<ObjectTypeSystem>();

	private List<ConstraintSet> constraintSets_ = new ArrayList<ConstraintSet>();

	private List knowledgeBase_ = new ArrayList();

	/**
	 * This method adds an object to the knowledge base. The previously
	 * installed ObjectTypeSystems and ObjectTypes are used to determine the
	 * type of the object.
	 * 
	 * @param obj
	 *            the object to add to the knowledge base
	 */
	public void add(Object obj) {

		List<ObjectType> types = findObjectTypes(obj);
		for (ObjectType type : types) {
			installType(type);
		}

		if (knowledgeManagers_.size() > 0) {
			ObjectView view = getObjectView(obj);

			if (view == null)
				throw new UnsupportedObjectType(
						"Could not find an ObjectView for the object " + obj);
			else
				objectViewLookup_.put(obj, view);

			for (KBManager mgr : knowledgeManagers_) {
				mgr.add(obj, view);
			}
		}

		for (ObjectType type : types) {
			if(!getObjectTypes(obj).contains(type))
				getObjectTypes(obj).add(type);
			if(!getObjectsByType(type).contains(obj))
				getObjectsByType(type).add(obj);
		}

		if (!knowledgeBase_.contains(obj))
			knowledgeBase_.add(obj);
	}

	/**
	 * This method delegates to the installed ObjectViews and ObjectTypeSystems
	 * to find an ObjectView of the specified Object.
	 * 
	 * @param obj
	 *            the object to find a view for
	 * @return the view for the object
	 */
	public ObjectView getObjectView(Object obj) {
		ObjectView view = null;
		for (ObjectView v : objectViews_) {
			if (v.supports(obj)) {
				view = v;
				break;
			}
		}
		if (view == null) {
			for (ObjectTypeSystem ots : typeSystems_) {
				if (ots.supports(obj)) {
					view = ots.getObjectView(obj);
					if (view != null)
						break;
				}
			}
		}
		return view;
	}

	/**
	 * This method returns all objects that are instances of the specified
	 * object type. This method only fetches objects that are cached as the
	 * required type. The ObjectTypeSystems are not consulted.
	 * 
	 * @param t
	 *            the type of objects to find
	 * @return the set of objects of the specified type
	 */
	public List getObjectsByType(ObjectType t) {
		List objs = objectsByType_.get(t);
		if (objs == null) {
			objs = new ArrayList();
			objectsByType_.put(t, objs);
		}
		return objs;
	}

	/**
	 * Removes an object from the knowledge base.
	 * 
	 * @param obj
	 *            the object o remove
	 */
	public void remove(Object obj) {
		List<ObjectType> types = findObjectTypes(obj);
		for (ObjectType type : types) {
			getObjectsByType(type).remove(obj);
		}
		
		if (knowledgeManagers_.size() > 0) {
			ObjectView view = null;
			for (ObjectView v : objectViews_) {
				if (v.supports(obj)) {
					view = v;
					break;
				}
			}

			if (view != null) {

				for (KBManager mgr : knowledgeManagers_) {
					mgr.remove(obj, view);
				}
			}
		}

		objectViews_.remove(obj);
		objectTypes_.remove(obj);
		knowledgeBase_.remove(obj);
	}

	/**
	 * This method adds a Constraint to the knowledge base and associates it
	 * with the specified ObjectType and relationship. If a targettype ObjecType
	 * is provided, it is used as an intial filter to find candidate endpoints
	 * for the relationship.
	 * 
	 * @param type
	 *            the type to bind the constraint to
	 * @param relationship
	 *            the relationship to bind the constraint to
	 * @param targettype
	 *            the targettype to filter potential endpoints
	 * @param con
	 *            the constraint to add
	 */
	public void add(ObjectType type, Object relationship,
			ObjectType targettype, Constraint con) {

		ExecutableQuery eq = getQuery(con);

		List qs = getQueries(type, relationship);
		setType(type, relationship, targettype);
		qs.add(eq);
	}
	
	
	public void setType(ObjectType type, Object relationship, ObjectType targettype){
		List qs = getQueries(type, relationship);

		if (qs.size() > 0) {
			if (targettype != null) {
				((ExecutableTypedSourceQuery) qs.get(0)).getTypes().add(
						targettype);
			}
		} else {
			ExecutableTypedSourceQuery source = new ExecutableTypedSourceQuery(
					this);
			qs.add(source);
			if (targettype != null) {
				source.getTypes().add(targettype);
			}
		}

	}

	/**
	 * You probably don't want to call this method
	 * directly. It won't hurt anything..but you are
	 * better off using validTargets(...) instead.
	 * 
	 * This method constructs a query evaluator for the
	 * specified type, relationship, and Constraint.
	 * 
	 * @param type the type of source object
	 * @param rel the relationship to query for
	 * @param con the constraint to satisfy
	 * @return the query evaluator
	 */
	public ExecutableQuery getQuery(Constraint con) {
		QueryEvaluator eval = queryEvaluators_.get(con.getType());
		if (eval == null) {
			ConstraintEvaluator ceval = constraintEvaluators_
					.get(con.getType());
			if (ceval == null) {
				throw new UnsupportedConstraintType(
						"No ConstraintEvaluator is defined for the constraint type: "
								+ con.getType());
			} else {
				eval = new IteratorQueryEvaluator(this);
			}
		}

		ExecutableQuery eq = new ExecutableQuery(eval, con);
		return eq;
	}

	/**
	 * 
	 * This method returns the set of objects in the knowledge base that meet
	 * the constraints placed on the source object's (the obj parameter)
	 * specified relationship (the rel parameter). The knowledge base will
	 * automatically attempt use the appropriate ConstraintEvaluators for all
	 * constraint types added to the relationship. For example, Groovy, OCL, and
	 * Prolog constraints can all be applied to the same feature and Refresh
	 * will figure out the appropriate search strategy to derive the appropriate
	 * endpoints meeting all of the diverse constraints.
	 * 
	 * IMPORTANT:
	 * You must use the add(ObjectType, Object, ObjectType, Constraint) method
	 * in conjunction with this method. That is, you must have specified what
	 * the target type of the relationship you are finding endpoints for is.
	 * Otherwise, it will attempt to evaluate your constraint against all 
	 * elements in the KB regardless of type. This is fine if your constraints
	 * can be evaluated correctly against all types of objects in the KB. If
	 * you are using ConstraintSets without the add(ObjectType, Object, ObjectType, Constraint)
	 * method, you need to use the version of this method that takes
	 * and additional ObjectType for the targettype as a parameter.
	 * @see validTargets(Object, Object, ObjectType)
	 * 
	 * @param obj
	 *            the object that is the source of the relationship
	 * @param rel
	 *            the relationship that endpoints are being found for
	 */
	public List validTargets(Object obj, Object rel) {
		return validTargets(obj, rel, (List)null);
	}
	
	/**
	 * This method returns the set of objects in the knowledge base that meet
	 * the constraints placed on the source object's (the obj parameter)
	 * specified relationship (the rel parameter). The objects returned
	 * will all be of the specified targettype. This method is designed
	 * to be used with ConstraintSets. When Constraints are directly added
	 * to the 
	 * 
	 * The knowledge base will
	 * automatically attempt use the appropriate ConstraintEvaluators for all
	 * constraint types added to the relationship. For example, Groovy, OCL, and
	 * Prolog constraints can all be applied to the same feature and Refresh
	 * will figure out the appropriate search strategy to derive the appropriate
	 * endpoints meeting all of the diverse constraints.
	 * 
	 * @param obj
	 *            the object that is the source of the relationship
	 * @param rel
	 *            the relationship that endpoints are being found for
	 */
	public List validTargets(Object obj, Object rel, ObjectType targettype) {
		return validTargets(obj, rel, getObjectsByType(targettype));
	}

	/**
	 * This method returns the set of objects in the knowledge base that meet
	 * the constraints placed on the source object's (the obj parameter)
	 * specified relationship (the rel parameter). The knowledge base will
	 * automatically attempt use the appropriate ConstraintEvaluators for all
	 * constraint types added to the relationship. For example, Groovy, OCL, and
	 * Prolog constraints can all be applied to the same feature and Refresh
	 * will figure out the appropriate search strategy to derive the appropriate
	 * endpoints meeting all of the diverse constraints.
	 * 
	 * The valid targets must come from the initial "world" or list of potential
	 * targets specified.
	 * 
	 * @param obj
	 *            the object that is the source of the relationship
	 * @param rel
	 *            the relationship that endpoints are being found for
	 * @param world
	 *            the list of potential targets
	 */
	public List validTargets(Object obj, Object rel, List world) {
		List<ExecutableQuery> qs = collectQueries(obj, rel);
		for (ExecutableQuery q : qs) {
			world = q.run(obj, world);
		}
		
		List<Constraint> cons = getConstraintsFromConstraintSets(obj, rel);
		for(Constraint con : cons){
			if(con != null){
				ExecutableQuery query = getQuery(con);
				world = query.run(obj, world);
			}
		}
		
		return world;
	}

	public List<Constraint> getConstraintsFromConstraintSets(Object obj,
			Object rel) {
		List<ObjectType> types = getObjectTypes(obj);
		List<Constraint> cons = new ArrayList<Constraint>();
		for (ConstraintSet cs : constraintSets_) {
			for (ObjectType t : types)
				cons.addAll(cs.get(obj, t, rel));
		}
		return cons;
	}

	/**
	 * This method returns true if all of the constraints placed on the source
	 * object's (the obj parameter) specified relationship (the rel parameter)
	 * hold true when the endpoint of the relationship is the specified target
	 * (the trg parameter). The knowledge base will automatically attempt use
	 * the appropriate ConstraintEvaluators for all constraint types added to
	 * the relationship. For example, Groovy, OCL, and Prolog constraints can
	 * all be applied to the same feature and Refresh will figure out the
	 * appropriate evaluation strategy to check the constraints.
	 * 
	 * @param obj
	 *            the object that is the source of the relationship
	 * @param rel
	 *            the relationship that is being checked
	 * @param trg
	 *            the endpoint of the relationship
	 */
	public boolean isValid(Object obj, Object rel, Object trg) {
		List world = new ArrayList();
		world.add(trg);
		return validTargets(obj, rel, world).contains(trg);
	}
	
	/**
	 * This method returns an empty list if all of the constraints placed on the source
	 * object's (the obj parameter) specified relationship (the rel parameter)
	 * hold true when the endpoint of the relationship is the specified target
	 * (the trg parameter). If one or more constraints are violated, the set of
	 * violated constraints is returned. The knowledge base will automatically attempt use
	 * the appropriate ConstraintEvaluators for all constraint types added to
	 * the relationship. For example, Groovy, OCL, and Prolog constraints can
	 * all be applied to the same feature and Refresh will figure out the
	 * appropriate evaluation strategy to check the constraints.
	 * 
	 * @param obj
	 *            the object that is the source of the relationship
	 * @param rel
	 *            the relationship that is being checked
	 * @param trg
	 *            the endpoint of the relationship
	 */
	public List<Constraint> violatedConstraint(Object obj, Object rel, Object target) {
		List world = new ArrayList();
		world.add(target);
		List<ExecutableQuery> qs = collectQueries(obj, rel);
		for (ExecutableQuery q : qs) {
			world = q.run(obj, world);
		}
		
		List<Constraint> cons = getConstraintsFromConstraintSets(obj, rel);
		for(Constraint con : cons){
			ExecutableQuery query = getQuery(con);
			world = query.run(obj, world);
			if(!world.contains(target))
				return query.getConstraints();
		}
		
		return Collections.EMPTY_LIST;
	}

	/**
	 * You probably don't want to use this method directly....
	 * 
	 * If you do, it returns the ExecutableQueries associated with a specific
	 * ObjectType and relationship.
	 * 
	 * @param obj
	 *            the object type
	 * @param rel
	 *            the relationship
	 * @return
	 */
	public List<ExecutableQuery> collectQueries(Object obj, Object rel) {
		List<ObjectType> types = getObjectTypes(obj);
		if (types == null || types.size() == 0) {
			types = findObjectTypes(obj);
		}

		List<ExecutableQuery> queries = new ArrayList<ExecutableQuery>();
		for (ObjectType t : types) {
			List<ExecutableQuery> tqueries = getQueries(t, rel);
			if (tqueries != null) {
				queries.addAll(tqueries);
			}
		}

		return queries;
	}

	/**
	 * This method uses all installed ObjectTypes and ObjectTypeSystems to find
	 * ObjectTypes that the provided object is an instance of.
	 * 
	 * @param obj
	 *            the object to find types for
	 * @return the set of types for the object
	 */
	public List<ObjectType> findObjectTypes(Object obj) {
		ArrayList<ObjectType> result = new ArrayList<ObjectType>();
		for (ObjectType type : objectTypes_) {
			if (type.instanceOf(obj)) {
				result.add(type);
			}
		}
		for (ObjectTypeSystem ots : typeSystems_) {
			if (ots.supports(obj)) {
				ObjectType type = ots.getObjectType(obj);
				if (type != null && !result.contains(type))
					result.add(type);
			}
		}

		return result;
	}

	/**
	 * Returns the ConstraintEvaluator that is registered to handle the
	 * specified type of constraint.
	 */
	public ConstraintEvaluator getConstraintEvaluator(String type) {
		return constraintEvaluators_.get(type);
	}

	/**
	 * This method is similar to findObjectTypes but does not consult the
	 * ObjectTypes and ObjectTypeSystems. Instead, this method consults a type
	 * cache and returns any previously known types for the object.
	 * 
	 * @param obj
	 *            the object of ind types for
	 * @return the types for the object
	 */
	public List<ObjectType> getObjectTypes(Object obj) {
		List<ObjectType> types = objectTypeLookup_.get(obj);
		if (types == null) {
			types = new ArrayList<ObjectType>();
			objectTypeLookup_.put(obj, types);
		}
		return types;
	}

	/**
	 * This method returns the queries associated with a specified ObjectType
	 * and relationship.
	 * 
	 * @param type
	 *            the type to find queries for
	 * @param relation
	 *            the relationship to find queries for
	 * @return the list of queries
	 */
	public List<ExecutableQuery> getQueries(ObjectType type, Object relation) {
		List<ExecutableQuery> queries = getQueryMap(type).get(relation);
		if (queries == null) {
			queries = new ArrayList<ExecutableQuery>();
			getQueryMap(type).put(relation, queries);
		}
		return queries;
	}

	/**
	 * You probably don't want to use this method directly...
	 * 
	 * If you do, this method returns a map that can be used to lookup the
	 * queries associated with each relationship of an ObjectType.
	 * 
	 * @param type
	 *            the type to lookup queries for
	 * @return the query map
	 */
	public Map<Object, List<ExecutableQuery>> getQueryMap(ObjectType type) {
		Map<Object, List<ExecutableQuery>> map = queryLookup_.get(type);
		if (map == null) {
			map = new HashMap<Object, List<ExecutableQuery>>();
			queryLookup_.put(type, map);
		}
		return map;
	}

	/**
	 * This method adds the specified ObjectType and any of its super types to
	 * the knowledge base that are not already installed.
	 * 
	 * @param type
	 *            the type to install
	 */
	public void installType(ObjectType type) {
		if (!getObjectTypes().contains(type)) {
			getObjectTypes().add(type);
			if (type.getSuperTypes() != null) {
				for (ObjectType t : type.getSuperTypes()) {
					if (!getObjectTypes().contains(t)) {
						getObjectTypes().add(t);
					}
				}
			}
		}
	}

	/**
	 * @return the constraintEvaluators
	 */
	public Map<String, ConstraintEvaluator> getConstraintEvaluators() {
		return constraintEvaluators_;
	}

	/**
	 * @param constraintEvaluators
	 *            the constraintEvaluators to set
	 */
	public void setConstraintEvaluators(
			Map<String, ConstraintEvaluator> constraintEvaluators) {
		constraintEvaluators_ = constraintEvaluators;
	}

	/**
	 * @return the knowledgeBase
	 */
	public List getKnowledgeBase() {
		return knowledgeBase_;
	}

	/**
	 * @param knowledgeBase
	 *            the knowledgeBase to set
	 */
	public void setKnowledgeBase(List knowledgeBase) {
		knowledgeBase_ = knowledgeBase;
	}

	/**
	 * @return the knowledgeManagers
	 */
	public List<KBManager> getKnowledgeManagers() {
		return knowledgeManagers_;
	}

	/**
	 * @param knowledgeManagers
	 *            the knowledgeManagers to set
	 */
	public void setKnowledgeManagers(List<KBManager> knowledgeManagers) {
		knowledgeManagers_ = knowledgeManagers;
	}

	/**
	 * @return the objectTypeLookup
	 */
	public Map<Object, List<ObjectType>> getObjectTypeLookup() {
		return objectTypeLookup_;
	}

	/**
	 * @param objectTypeLookup
	 *            the objectTypeLookup to set
	 */
	public void setObjectTypeLookup(
			Map<Object, List<ObjectType>> objectTypeLookup) {
		objectTypeLookup_ = objectTypeLookup;
	}

	/**
	 * @return the objectTypes
	 */
	public List<ObjectType> getObjectTypes() {
		return objectTypes_;
	}

	/**
	 * @param objectTypes
	 *            the objectTypes to set
	 */
	public void setObjectTypes(List<ObjectType> objectTypes) {
		objectTypes_ = objectTypes;
	}

	/**
	 * @return the objectViewLookup
	 */
	public Map<Object, ObjectView> getObjectViewLookup() {
		return objectViewLookup_;
	}

	/**
	 * @param objectViewLookup
	 *            the objectViewLookup to set
	 */
	public void setObjectViewLookup(Map<Object, ObjectView> objectViewLookup) {
		objectViewLookup_ = objectViewLookup;
	}

	/**
	 * @return the objectViews
	 */
	public List<ObjectView> getObjectViews() {
		return objectViews_;
	}

	/**
	 * @param objectViews
	 *            the objectViews to set
	 */
	public void setObjectViews(List<ObjectView> objectViews) {
		objectViews_ = objectViews;
	}

	/**
	 * @return the queryEvaluators
	 */
	public Map<String, QueryEvaluator> getQueryEvaluators() {
		return queryEvaluators_;
	}

	/**
	 * @param queryEvaluators
	 *            the queryEvaluators to set
	 */
	public void setQueryEvaluators(Map<String, QueryEvaluator> queryEvaluators) {
		queryEvaluators_ = queryEvaluators;
	}

	/**
	 * @return the queryLookup
	 */
	public Map<ObjectType, Map<Object, List<ExecutableQuery>>> getQueryLookup() {
		return queryLookup_;
	}

	/**
	 * @param queryLookup
	 *            the queryLookup to set
	 */
	public void setQueryLookup(
			Map<ObjectType, Map<Object, List<ExecutableQuery>>> queryLookup) {
		queryLookup_ = queryLookup;
	}

	public List<ObjectTypeSystem> getTypeSystems() {
		return typeSystems_;
	}

	public void setTypeSystems(List<ObjectTypeSystem> typeSystems) {
		typeSystems_ = typeSystems;
	}

	public List<ConstraintSet> getConstraintSets() {
		return constraintSets_;
	}

	public void setConstraintSets(List<ConstraintSet> constraintSets) {
		constraintSets_ = constraintSets;
	}

}
